import abc
import shutil
import venv  # type: ignore
import pkg_resources

from amino import Path, IO, do, Maybe, _, L, List, Future, Boolean, Map, Either, Right
from amino.util.string import ToStr
from amino.boolean import true, false

from ribosome import Job
from ribosome.process import JobClient, Result

from chromatin.plugin import RpluginSpec

from chromatin.logging import Logging
from chromatin.venv import Venv, PluginVenv


class VenvState(ToStr):

    def __init__(self, plugin: RpluginSpec) -> None:
        self.plugin = plugin

    def _arg_desc(self) -> List[str]:
        return List(str(self.plugin))


class VenvExistent(VenvState):

    def __init__(self, plugin, venv) -> None:
        super().__init__(plugin)
        self.venv = venv


class VenvAbsent(VenvState):
    pass


@do
def remove_dir(dir: Path) -> IO[None]:
    exists = yield IO.delay(dir.exists)
    yield IO.delay(shutil.rmtree, dir) if exists else IO.pure(None)


@do
def build(dir: Path, plugin: RpluginSpec) -> IO[Venv]:
    builder = venv.EnvBuilder(system_site_packages=False, with_pip=True)
    context = yield IO.delay(builder.ensure_directories, str(dir))
    yield IO.delay(builder.create_configuration, context)
    yield IO.delay(builder.setup_python, context)
    yield IO.delay(builder._setup_pip, context)
    yield IO.delay(builder.setup_scripts, context)
    yield IO.delay(builder.post_setup, context)
    yield IO.pure(Venv.from_ns(dir, plugin, context))


def cons_venv(dir: Path, plugin: RpluginSpec) -> Venv:
    builder = venv.EnvBuilder(system_site_packages=False, with_pip=True)
    context = builder.ensure_directories(str(dir))
    return Venv.from_ns(dir, plugin, context)


class PackageState(ToStr):

    def __init__(self, venv: Venv) -> None:
        self.venv = venv

    @abc.abstractproperty
    def exists(self) -> Boolean:
        ...


class PackageExistent(PackageState):

    def __init__(self, venv: Venv, dist: pkg_resources.Distribution) -> None:
        self.venv = venv
        self.dist = dist

    def _arg_desc(self) -> List[str]:
        return List(str(self.venv), str(self.dist))

    @property
    def exists(self) -> Boolean:
        return true


class PackageAbsent(PackageState):

    def _arg_desc(self) -> List[str]:
        return List()

    @property
    def exists(self) -> Boolean:
        return false


def package_state(venv: Venv, req: str) -> PackageState:
    ws = pkg_resources.WorkingSet([venv.site])
    req = pkg_resources.Requirement(req)
    return Maybe.check(ws.by_key.get(req.key)) / L(PackageExistent)(venv, _) | PackageAbsent(venv)


def package_state_main(venv: Venv) -> PackageState:
    return package_state(venv, venv.name)


class VenvFacade(Logging):

    def __init__(self, dir: Path) -> None:
        self.dir = dir

    def check(self, plugin: RpluginSpec) -> VenvState:
        dir = self.dir / plugin.name
        return (
            VenvExistent(plugin, cons_venv(dir, plugin))
            if dir.exists() else
            VenvAbsent(plugin)
        )

    @do
    def bootstrap(self, plugin: RpluginSpec) -> IO[Venv]:
        venv_dir = self.dir / plugin.name
        self.log.debug(f'bootstrapping {plugin} in {venv_dir}')
        yield remove_dir(venv_dir)
        yield build(venv_dir, plugin)

    def package_state(self, venv: Venv) -> PackageState:
        return package_state_main(venv)

    def package_installed(self, venv: Venv) -> Boolean:
        return self.package_state(venv).exists

    @do
    def install(self, pvenv: PluginVenv) -> Either[str, Future[Result]]:
        venv = pvenv.venv
        self.log.debug(f'installing {venv}')
        bin_path = yield venv.bin_path
        pip_bin = bin_path / 'pip'
        args = ['install', '-U', '--no-cache', pvenv.req]
        job = Job(
            client=JobClient(cwd=Path.cwd(), name=f'pip install -U {pvenv.req}'),
            exe=str(pip_bin),
            args=args,
            loop=None,
            kw=Map(env=dict()),
        )
        yield Right(job)

__all__ = ('VenvFacade', 'VenvState', 'VenvExistent', 'VenvAbsent')



from ribosome.machine import message, json_message

Start = message('Start')
ReadConf = message('ReadConf')
AddPlugin = json_message('AddPlugin', 'spec')
ShowPlugins = message('ShowPlugins')
SetupPlugins = message('SetupPlugins')
SetupVenvs = message('SetupVenvs')
PostSetup = message('PostSetup')
InstallMissing = message('InstallMissing')
AddVenv = message('AddVenv', 'venv')
IsInstalled = message('Installed', 'venv')
Installed = message('Installed', 'venv')
Updated = message('Updated', 'venv')
Activate = message('Activate', varargs='plugins')
Deactivate = message('Deactivate', varargs='plugins')
Activated = message('Activated', 'venv')
Deactivated = message('Deactivated', 'venv')
AlreadyActive = message('AlreadyActive', 'venv')
ActivationComplete = message('ActivationComplete')
InitializationComplete = message('InitializationComplete')
UpdatePlugins = message('UpdatePlugins', varargs='plugins')
Reboot = message('Reboot', varargs='plugins')
DefinedHandlers = message('DefinedHandlers', 'venv', 'handlers')

__all__ = ('ReadConf', 'Start', 'AddPlugin', 'ShowPlugins', 'SetupPlugins', 'SetupVenvs', 'InstallMissing',
           'AddVenv', 'IsInstalled', 'Installed', 'Updated', 'Activate', 'Deactivate', 'Activated', 'Deactivated',
           'PostSetup', 'AlreadyActive', 'ActivationComplete', 'InitializationComplete', 'UpdatePlugins', 'Reboot',
           'DefinedHandlers')



from ribosome.record import Record, str_field, uuid_field

from amino import List, Either, Map


class RpluginSpec(Record):
    uuid = uuid_field()
    name = str_field()
    spec = str_field()

    @staticmethod
    def from_config(data: dict) -> Either[str, 'RpluginSpec']:
        m = Map(data)
        def create(spec: str) -> RpluginSpec:
            name = m.lift('name') | spec
            return RpluginSpec(name=str(name), spec=str(spec))
        return m.lift('spec').to_either(f'plugin data {data} missing attribute `spec`').map(create)

    def _arg_desc(self) -> List[str]:
        return List(self.name, self.spec)

__all__ = ('RpluginSpec',)



from amino import List, Just, Maybe

from ribosome import NvimFacade
from ribosome.machine.state import UnloopedRootMachine

from chromatin.env import Env
from chromatin.logging import Logging


class Chromatin(UnloopedRootMachine, Logging):
    _data_type = Env

    def __init__(self, vim: NvimFacade, plugins: List[str]) -> None:
        core = 'chromatin.plugins.core'
        UnloopedRootMachine.__init__(self, vim, plugins.cons(core))

    @property
    def init(self):
        return Env(vim_facade=Just(self.vim))

    @property
    def title(self):
        return 'chromatin'

__all__ = ('Chromatin',)



from types import SimpleNamespace
from typing import Any

from amino import Path, List, Try

from ribosome.record import Record, field, either_field, path_field, int_field, str_field

from chromatin.plugin import RpluginSpec


class Venv(Record):
    name = str_field()
    dir = path_field()
    python_executable = either_field(Path, factory=Path)
    bin_path = either_field(Path, factory=Path)

    @staticmethod
    def from_ns(dir: Path, plugin: RpluginSpec, context: SimpleNamespace) -> 'Venv':
        exe = Try(lambda: context.env_exe) / Path
        bin_path = Try(lambda: context.bin_path) / Path
        return Venv(name=plugin.name, dir=dir, python_executable=exe, bin_path=bin_path)

    def _arg_desc(self) -> List[str]:
        return List(str(self.dir), str(self.name))

    @property
    def site(self) -> Path:
        return self.dir / 'lib' / 'python3.6' / 'site-packages'

    @property
    def plugin_path(self) -> Path:
        return self.site / self.name / 'nvim_plugin.py'


class PluginVenv(Record):
    venv = field(Venv)
    plugin = field(RpluginSpec)

    @property
    def req(self) -> str:
        return self.plugin.spec

    @property
    def name(self) -> str:
        return self.plugin.name

    def _str_extra(self) -> List[Any]:
        return List(self.venv, self.plugin)


class ActiveVenv(Record):
    venv = field(Venv)
    channel = int_field()
    pid = int_field()

    @property
    def name(self) -> str:
        return self.venv.name

    @property
    def _str_extra(self) -> List[Any]:
        return List(self.venv, self.channel, self.pid)

__all__ = ('Venv', 'ActiveVenv', 'PluginVenv')



import typing
from typing import Tuple

from ribosome.nvim import NvimIO

from amino import Path, do


def host_cmdline(python_exe: Path, plug: Path, debug: bool) -> typing.List[str]:
    debug_option = '' if debug else 'E'
    return [str(python_exe), f'-{debug_option}c', 'import neovim; neovim.start_host()', str(plug)]


@do
def start_host(python_exe: Path, plugin_path: Path, debug: bool=False) -> NvimIO[Tuple[int, int]]:
    cmdline = host_cmdline(python_exe, plugin_path, debug)
    channel = yield NvimIO.call('jobstart', cmdline, dict(rpc=True, on_stderr='ChromatinJobStderr'))
    pid = yield NvimIO.call('jobpid', channel)
    yield NvimIO.pure((channel, pid))


def stop_host(channel: int) -> NvimIO[None]:
    return NvimIO.call('jobstop', channel)

__all__ = ('start_host', 'stop_host')



import time
import threading

import neovim

from amino import List

from ribosome import NvimStatePlugin
from ribosome.request import msg_command, json_msg_command
from ribosome.nvim import NvimFacade

from chromatin.main import Chromatin
from chromatin.plugins.core.messages import (AddPlugin, ShowPlugins, Start, SetupPlugins, UpdatePlugins, Activate,
                                             Deactivate, Reboot)


@neovim.plugin
class ChromatinNvimPlugin(NvimStatePlugin, name='chromatin', prefix='crm'):

    def __init__(self, vim: neovim.api.Nvim) -> None:
        super().__init__(NvimFacade(vim, 'chromatin'))
        self.initialized = False
        self.chromatin: Chromatin = None

    @property
    def _default_plugins(self) -> List[str]:
        return List()

    def stage_1(self) -> None:
        plugins = self.vim.vars.pl('plugins') | self._default_plugins
        self.chromatin = Chromatin(self.vim.proxy, plugins)
        self.chromatin.start()
        self.chromatin.wait_for_running()
        self.chromatin.send(Start())

    def state(self) -> Chromatin:
        if self.chromatin is None and not self.initialized:
            self.initialized = True
            self.stage_1()
        self.wait_for_startup()
        return self.chromatin

    def wait_for_startup(self) -> bool:
        start = time.time()
        while self.chromatin is None and time.time() - start < 5:
            time.sleep(.1)
        if self.chromatin is not None:
            self.chromatin.wait_for_running()
            time.sleep(.1)
        return self.chromatin is not None

    @json_msg_command(AddPlugin)
    def cram(self) -> None:
        pass

    @msg_command(ShowPlugins)
    def crm_show_plugins(self) -> None:
        pass

    @msg_command(SetupPlugins)
    def crm_setup_plugins(self) -> None:
        pass

    @msg_command(Activate)
    def crm_activate(self) -> None:
        pass

    @msg_command(Deactivate)
    def crm_deactivate(self) -> None:
        pass

    @msg_command(Reboot)
    def crm_reboot(self) -> None:
        pass

    @msg_command(UpdatePlugins)
    def crm_update(self) -> None:
        pass

    @neovim.autocmd('VimEnter')
    def vim_enter(self):
        threading.Thread(target=self.state).start()


__all__ = ('ChromatinNvimPlugin',)



from amino import Path, List
from amino.options import EnvOption

xdg_cache_home = EnvOption('XDG_CACHE_HOME')


def create_venv_dir_error(dir: Path) -> str:
    return f'could not create venv dir `{dir}`'


def no_plugins_match_for_update(names: List[str]) -> str:
    return f'no plugins match for update: {names.join_comma}'


def no_plugins_match_for_activation(names: List[str]) -> str:
    return f'no plugins match for activation: {names.join_comma}'


def no_plugins_match_for_deactivation(names: List[str]) -> str:
    return f'no plugins match for deactivation: {names.join_comma}'


def installed_plugin(name: str) -> str:
    return f'installed plugin \'{name}\''


def updated_plugin(name: str) -> str:
    return f'updated plugin \'{name}\''

__all__ = ('xdg_cache_home', 'xdg_cache_home_env_var', 'create_venv_dir_error', 'installed_plugin',
           'updated_plugin', 'no_plugins_match_for_activation', 'no_plugins_match_for_deactivation')



import sys
import inspect

import neovim

from amino.logging import TEST
from amino import Path

from ribosome.logging import nvim_logging, ribo_log
from ribosome.nvim import NvimFacade
from ribosome.rpc import rpc_handlers, define_handlers

from chromatin.nvim_plugin import ChromatinNvimPlugin
from chromatin.host import start_host


def run() -> None:
    try:
        installed = sys.argv[2] == '1'
        nvim_native = neovim.attach('stdio')
        nvim = NvimFacade(nvim_native, 'define_handlers').proxy
        nvim_logging(nvim, TEST)
        python_exe = Path(sys.argv[1])
        plugin_path = Path(inspect.getfile(ChromatinNvimPlugin))
        channel, pid = start_host(python_exe, plugin_path, True).attempt(nvim).get_or_raise
        handlers = rpc_handlers(ChromatinNvimPlugin)
        define_handlers(channel, handlers, 'chromatin', str(plugin_path)).attempt(nvim).get_or_raise
        if installed:
            ribo_log.info('chromatin initialized. installing plugins...')
        nvim.cmd('ChromatinStage1')
        return 0
    except Exception as e:
        ribo_log.error(e)
        return 1

__all__ = ('run',)



from typing import Any, Generator

from ribosome.data import Data
from ribosome.record import dfield, list_field, map_field, maybe_field
from ribosome.nvim import NvimFacade, AsyncVimProxy
from ribosome.rpc import DefinedHandler

from amino import List, _, Either, Path, Try, Right, do, Boolean, Map, __
from amino.boolean import true

from chromatin.logging import Logging
from chromatin.plugin import RpluginSpec
from chromatin.venv import Venv, ActiveVenv, PluginVenv
from chromatin.venvs import VenvFacade
from chromatin.util.resources import xdg_cache_home, create_venv_dir_error


@do
def _default_venv_dir() -> Generator[Either[str, Path], Any, None]:
    xdg_cache_path = xdg_cache_home.value / Path | (Path.home() / '.cache')
    venv_dir = xdg_cache_path / 'chromatin' / 'venvs'
    yield Try(venv_dir.mkdir, parents=True, exist_ok=True).lmap(lambda a: create_venv_dir_error(venv_dir))
    yield Right(venv_dir)


class Env(Logging, Data):
    vim_facade = maybe_field((NvimFacade, AsyncVimProxy))
    initialized = dfield(False)
    plugins = list_field(RpluginSpec)
    venvs = map_field()
    installed = list_field(Venv)
    active = list_field(ActiveVenv)
    uninitialized = list_field(ActiveVenv)
    handlers = map_field()

    def add_plugin(self, name: str, spec: str) -> 'Env':
        return self.append1.plugins(RpluginSpec(name=name, spec=spec))

    def add_plugins(self, plugins: List[RpluginSpec]) -> 'Env':
        return self.append.plugins(plugins)

    @property
    def show_plugins(self) -> List[str]:
        def format(plug: RpluginSpec) -> str:
            return plug.spec
        return self.plugins.map(format).cons('Configured plugins:')

    def add_venv(self, venv: Venv) -> 'Env':
        return self.modder.venvs(_ + (venv.name, venv))

    def add_installed(self, venv: Venv) -> 'Env':
        return self.append1.installed(venv)

    def plugin_by_name(self, name: str) -> Either[str, RpluginSpec]:
        return self.plugins.find(lambda a: a.name == name)

    @do
    def plugin_venv(self, venv: Venv) -> Either[str, PluginVenv]:
        plugin = yield self.plugin_by_name(venv.name)
        yield Right(PluginVenv(venv=venv, plugin=plugin))

    def missing_in(self, venvs: VenvFacade) -> List[Venv]:
        return self.venvs.v.filter_not(venvs.package_installed)

    @property
    @do
    def missing(self) -> Either[str, List[Venv]]:
        venv_facade = yield self.venv_facade
        yield Right(self.missing_in(venv_facade))

    @property
    def _str_extra(self) -> List[Any]:
        return List(self.plugins, self.venvs)

    @property
    def venv_dir(self) -> Either[str, Path]:
        return self.vim.vars.ppath('venv_dir').o(_default_venv_dir)

    @property
    def venv_facade(self) -> Either[str, VenvFacade]:
        return self.venv_dir / VenvFacade

    @property
    def want_init(self) -> Boolean:
        return self.vim.vars.pb('autostart') | true

    @property
    def autoreboot(self) -> Boolean:
        return self.vim.vars.pb('autoreboot') | true

    @property
    def active_venvs(self) -> List[Venv]:
        return self.active / _.venv

    @property
    def inactive(self) -> List[ActiveVenv]:
        return self.installed.remove_all(self.active_venvs)

    def host_started(self, venv: ActiveVenv) -> 'Env':
        return self.append1.active(venv).append1.uninitialized(venv)

    def initialization_complete(self) -> 'Env':
        return self.setter.uninitialized(List())

    def deactivate_venv(self, venv: ActiveVenv) -> 'Env':
        return self.modder.active(__.without(venv))

    def installed_by_name(self, names: List[str]) -> List[Venv]:
        return self.installed.filter(lambda v: v.name in names)

    def active_by_name(self, names: List[str]) -> List[str]:
        return self.active.filter(lambda v: v.name in names)

    def to_map(self) -> Map[str, Any]:
        return super().to_map() - 'vim_facade'

    @property
    def vim(self) -> NvimFacade:
        return self.vim_facade | (lambda: NvimFacade(None, 'no vim was set in `Env`'))

    def add_handlers(self, venv: Venv, handlers: List[DefinedHandler]) -> 'Env':
        return self.modder.handlers(_ + (venv.name, handlers))

    def handlers_for(self, plugin: str) -> Either[str, List[DefinedHandler]]:
        return self.handlers.lift(plugin).to_either(f'no handlers defined for {plugin}')

__all__ = ('Env',)



from typing import Generator, TypeVar, Any

from lenses import Lens, lens

from amino.state import State, EitherState, StateT
from amino import __, do, _, Either, L, Just, List, Right, Boolean, Lists, Maybe, curried, Nothing, Path, Id
from amino.util.string import camelcaseify, camelcase
from amino.boolean import true, false
from amino.list import Nil

from ribosome.machine import Message
from ribosome.machine.base import io, RunIOsParallel, SubProcessSync
from ribosome.machine.transition import Error
from ribosome.machine import trans
from ribosome.nvim import NvimIO
from ribosome.machine.messages import Info, RunNvimIO
from ribosome.rpc import define_handler, RpcHandlerSpec, DefinedHandler, define_handlers
from ribosome.process import Result

from chromatin.state import ChromatinTransitions, ChromatinComponent
from chromatin.plugins.core.messages import (AddPlugin, ShowPlugins, Start, SetupPlugins, SetupVenvs, InstallMissing,
                                             AddVenv, IsInstalled, Activated, PostSetup, Installed, UpdatePlugins,
                                             Updated, Reboot, Activate, AlreadyActive, ReadConf, Deactivate,
                                             Deactivated, DefinedHandlers, ActivationComplete, InitializationComplete)
from chromatin.venvs import VenvExistent
from chromatin.env import Env
from chromatin.venv import Venv, ActiveVenv, PluginVenv
from chromatin.logging import Logging
from chromatin.host import start_host, stop_host
from chromatin.util import resources
from chromatin.plugin import RpluginSpec

A = TypeVar('A')
STE = StateT[Either, Env, A]
STI = StateT[Id, Env, A]
ESL = Generator[STE, Any, None]
ISL = Generator[STI, Any, None]


class PluginFunctions(Logging):

    @do
    def setup_venvs(self) -> ESL:
        '''check whether a venv exists for each plugin in the self.env.
        for those that don't, create self.venvs in `g:chromatin_venv_dir`.
        '''
        venv_facade = yield EitherState.inspect_f(_.venv_facade)
        plugins = yield EitherState.inspect(_.plugins)
        jobs = plugins / venv_facade.check
        existent, absent = jobs.split_type(VenvExistent)
        ios = absent / _.plugin / venv_facade.bootstrap / __.map(AddVenv).map(_.pub)
        yield EitherState.pure(existent.map(_.venv).map(IsInstalled).cat(RunIOsParallel(ios)))

    @do
    def install_plugins(self, venvs: List[Venv], update: Boolean) -> ESL:
        '''run subprocesses in sequence that install packages into their venvs using pip.
        cannot be run in parallel as they seem to deadlock.
        '''
        Done = update.cata(Updated, Installed)
        venv_facade = yield EitherState.inspect_f(_.venv_facade)
        def trans_result(venv: Venv, result: Result) -> Maybe[Message]:
            return Just((Done(venv) if result.success else Error(result.err)).pub)
        @do
        def install_proc(pvenv: PluginVenv) -> Generator[Either[str, SubProcessSync], Any, None]:
            job = yield venv_facade.install(pvenv)
            yield Right(SubProcessSync(job, curried(trans_result)(pvenv.venv)))
        pvenvs = yield EitherState.inspect_f(lambda env: venvs.traverse(env.plugin_venv, Right))
        yield EitherState.lift(pvenvs.traverse(install_proc, Either))

    @do
    def install_missing(self) -> ESL:
        missing = yield EitherState.inspect_f(_.missing)
        yield self.install_plugins(missing, false)

    @do
    def update_plugins(self, plugins: List[str]) -> ESL:
        getter = _.installed if plugins.empty else __.installed_by_name(plugins)
        venvs = yield EitherState.inspect(getter)
        yield (
            EitherState.pure(List(Error(resources.no_plugins_match_for_update(plugins))))
            if venvs.empty else
            self.install_plugins(venvs, true)
        )

    @do
    def start_host(self, venv: Venv, python_exe: Path) -> Generator[NvimIO[Activated], Any, None]:
        channel, pid = yield start_host(python_exe, venv.plugin_path)
        yield NvimIO.pure(List(Activated(ActiveVenv(venv=venv, channel=channel, pid=pid))))

    @do
    def activate_venv(self, venv: Venv) -> Generator[Either[str, NvimIO[Activated]], Any, None]:
        python_exe = yield venv.python_executable
        yield Right(RunNvimIO(self.start_host(venv, python_exe)))

    @do
    def activate_multi(self, venvs: List[Venv]) -> ESL:
        active = yield EitherState.inspect(_.active_venvs)
        already_active, inactive = venvs.split(active.contains)
        aa_msgs = already_active / AlreadyActive
        ios = inactive / self.activate_venv / __.value_or(Error)
        yield EitherState.pure((aa_msgs + ios).cat(ActivationComplete()))

    @do
    def activate_by_names(self, plugins: List[str]) -> ESL:
        getter = _.installed if plugins.empty else __.installed_by_name(plugins)
        venvs = yield EitherState.inspect(getter)
        yield (
            EitherState.pure(List(Error(resources.no_plugins_match_for_activation(plugins))))
            if venvs.empty else
            self.activate_multi(venvs)
        )

    def activate_all(self) -> STE:
        return self.activate_by_names(List())

    def deactivate_venv(self, venv: ActiveVenv) -> State[Env, RunNvimIO]:
        def undef(spec: RpcHandlerSpec) -> NvimIO[str]:
            return NvimIO.cmd(spec.undef_cmdline, verbose=True)
        @do
        def run(env: Env) -> Generator[NvimIO[Any], Any, None]:
            handlers = (env.handlers_for(venv.name) | Nil) / _.spec
            yield stop_host(venv.channel)
            yield handlers.traverse(undef, NvimIO)
            yield NvimIO.pure(List(Deactivated(venv)))
        return State.inspect(run) / RunNvimIO

    def deactivate_multi(self, venvs: List[ActiveVenv]) -> State[Env, List[RunNvimIO]]:
        return venvs.traverse(self.deactivate_venv, State)

    @do
    def deactivate_by_names(self, plugins: List[str]) -> ISL:
        getter = _.active if plugins.empty else __.active_by_name(plugins)
        venvs = yield State.inspect(getter)
        yield (
            State.pure(List(Error(resources.no_plugins_match_for_deactivation(plugins))))
            if venvs.empty else
            self.deactivate_multi(venvs)
        )

    @do
    def activate_newly_installed(self) -> ESL:
        new = yield EitherState.inspect(_.inactive)
        yield self.activate_multi(new)

    @do
    def define_handlers(self, active_venv: ActiveVenv) -> Generator[NvimIO[List[DefinedHandler]], Any, None]:
        venv = active_venv.venv
        name = venv.name
        channel = active_venv.channel
        cname = camelcaseify(name)
        rpc_handlers_fun = f'{cname}RpcHandlers'
        handlers_spec = RpcHandlerSpec.fun(1, rpc_handlers_fun, dict())
        handler_rpc = yield define_handler(channel, handlers_spec, name, venv.plugin_path)
        result = yield NvimIO.call_once_defined(rpc_handlers_fun, timeout=3)
        handler_specs = Lists.wrap(result).flat_map(RpcHandlerSpec.decode)
        handlers = yield define_handlers(channel, handler_specs, name, venv.plugin_path)
        yield NvimIO.pure(handlers.cons(handler_rpc))

    @do
    def add_plugins(self, plugins: List[RpluginSpec]) -> ESL:
        yield EitherState.modify(__.add_plugins(plugins))
        init = yield EitherState.inspect(_.want_init)
        yield EitherState.pure(init.m(SetupPlugins()))

    @do
    def read_conf(self) -> ESL:
        vim = yield EitherState.inspect(_.vim)
        plugins = vim.vars.pl('rplugins').flat_map(__.traverse(RpluginSpec.from_config, Either))
        yield (
            EitherState.pure(Nothing)
            if plugins.exists(_.empty) else
            plugins.map(self.add_plugins).value_or(lambda a: EitherState.pure(Just(Error(a))))
        )


class CoreTransitions(ChromatinTransitions):

    @property
    def funcs(self) -> PluginFunctions:
        return PluginFunctions()

    @trans.multi(Start)
    def stage_i(self) -> List[Message]:
        return List(io(__.vars.set_p('started', True)), io(__.vars.ensure_p('rplugins', [])), ReadConf().at(0.6).pub)

    @trans.one(ReadConf, trans.est, trans.m)
    def read_conf(self) -> EitherState[Env, Maybe[Message]]:
        return self.funcs.read_conf()

    @trans.one(AddPlugin, trans.st, trans.m)
    @do
    def add_plugin(self) -> State[Env, Maybe[Message]]:
        init = yield State.inspect(_.want_init)
        spec = self.msg.spec
        name = self.msg.options.get('name') | spec
        yield State.modify(__.add_plugin(name, spec))
        yield State.pure(init.m(SetupPlugins().at(.75)))

    @trans.one(ShowPlugins, trans.st)
    @do
    def show_plugins(self) -> State[Env, Message]:
        venv_dir = yield State.inspect(_.venv_dir)
        venv_dir_msg = f'virtualenv dir: {venv_dir.value}'
        yield State.pure(Info(self.data.show_plugins.cons(venv_dir_msg).join_lines))

    @trans.multi(SetupPlugins)
    def setup_plugins(self) -> List[Message]:
        return List(SetupVenvs(), InstallMissing().at(.9).pub, PostSetup().at(.95).pub)

    @trans.multi(SetupVenvs, trans.est)
    def setup_venvs(self) -> EitherState[Env, List[Message]]:
        return self.funcs.setup_venvs()

    @trans.multi(InstallMissing, trans.est)
    def install_missing(self) -> EitherState[Env, List[Message]]:
        return self.funcs.install_missing()

    @trans.multi(Installed, trans.st)
    def installed(self) -> State[Env, List[Message]]:
        venv = self.msg.venv
        return State.pure(List(IsInstalled(venv), Info(resources.installed_plugin(venv.name))))

    @trans.multi(Updated, trans.st)
    @do
    def updated(self) -> State[Env, List[Message]]:
        autoreboot = yield State.inspect(_.autoreboot)
        venv = self.msg.venv
        reboot = autoreboot.l(Reboot(venv.name))
        yield State.pure(reboot.cons(Info(resources.updated_plugin(venv.name))))

    @trans.unit(IsInstalled, trans.st)
    def is_installed(self) -> State[Env, None]:
        return State.modify(__.add_installed(self.msg.venv))

    @trans.multi(PostSetup, trans.est)
    def post_setup(self) -> EitherState[Env, List[Message]]:
        return self.funcs.activate_newly_installed()

    @trans.unit(AddVenv, trans.st)
    def add_venv(self) -> State[Env, None]:
        return State.modify(__.add_venv(self.msg.venv))

    @trans.multi(Activate, trans.est)
    def activate(self) -> EitherState[Env, List[Message]]:
        return self.funcs.activate_by_names(self.msg.plugins)

    @trans.multi(Deactivate, trans.st)
    def deactivate(self) -> State[Env, List[Message]]:
        return self.funcs.deactivate_by_names(self.msg.plugins)

    @trans.multi(Reboot)
    def reboot(self) -> List[Message]:
        plugins = self.msg.plugins
        return List(Deactivate(*plugins), Activate(*plugins).at(.75).pub)

    @trans.one(Activated, trans.st, trans.nio)
    @do
    def activated(self) -> State[Env, NvimIO[Message]]:
        self.log.debug(f'activated {self.msg.venv}')
        active_venv = self.msg.venv
        venv = active_venv.venv
        @do
        def io() -> NvimIO[DefinedHandlers]:
            handlers = yield self.funcs.define_handlers(active_venv)
            yield self.vim.runtime(f'chromatin/{venv.name}/*')
            yield NvimIO.pure(DefinedHandlers(venv, handlers))
        yield State.modify(__.host_started(active_venv))
        yield State.pure(io())

    @trans.one(ActivationComplete, trans.st, trans.nio)
    @do
    def activation_complete(self) -> State[Env, NvimIO[None]]:
        venvs = yield State.inspect(_.uninitialized)
        prefixes = venvs / _.name / camelcase
        def stage(num: int) -> NvimIO[None]:
            return prefixes.traverse(lambda a: NvimIO.cmd_sync(f'{a}Stage{num}'), NvimIO)
        yield State.pure(Lists.range(1, 5).traverse(stage, NvimIO).replace(InitializationComplete()))

    @trans.unit(InitializationComplete, trans.st)
    def initialization_complete(self) -> State[Env, None]:
        return State.modify(__.initialization_complete())

    @trans.unit(Deactivated, trans.st)
    def deactivated(self) -> None:
        self.log.debug(f'deactivated {self.msg.venv}')
        return State.modify(__.deactivate_venv(self.msg.venv))

    @trans.unit(DefinedHandlers, trans.st)
    def defined_handlers(self) -> State[Env, None]:
        return State.modify(__.add_handlers(self.msg.venv, self.msg.handlers))

    @trans.multi(UpdatePlugins, trans.est)
    def update_plugins(self) -> State[Env, List[Message]]:
        return self.funcs.update_plugins(self.msg.plugins)

    @trans.unit(AlreadyActive)
    def already_active(self) -> None:
        pass

    def state_lens(self, tpe: str, name: str) -> Either[str, Lens]:
        return (
            State.inspect(lambda s: s.plugins.index_where(lambda a: a.name == name) / (lambda i: lens(s).plugins[i]))
            if tpe == 'vim_plugin' else
            State.pure(Nothing)
        )


class Plugin(ChromatinComponent):
    Transitions = CoreTransitions

__all__ = ('Plugin',)



import amino.logging
import ribosome.logging
from ribosome.logging import ribosome_logger

from amino.lazy import lazy


log = chromatin_root_logger = ribosome_logger('chromatin')


def chromatin_logger(name: str):
    return chromatin_root_logger.getChild(name)


class Logging(ribosome.logging.Logging):

    @lazy
    def _log(self) -> amino.logging.Logger:
        return chromatin_logger(self.__class__.__name__)

__all__ = ('chromatin_logger', 'Logging')



from ribosome import NvimFacade
from ribosome.nvim import HasNvim
from ribosome.machine.state import SubMachine, SubTransitions
from ribosome.machine.base import MachineBase

from chromatin.logging import Logging


class ChromatinComponent(SubMachine, HasNvim, Logging):

    def __init__(self, vim: NvimFacade, parent=None, title=None) -> None:
        MachineBase.__init__(self, parent, title=title)
        HasNvim.__init__(self, vim)


class ChromatinTransitions(SubTransitions, HasNvim, Logging):

    def __init__(self, machine, *a, **kw):
        SubTransitions.__init__(self, machine, *a, **kw)
        HasNvim.__init__(self, machine.vim)

__all__ = ('ChromatinComponent', 'ChromatinTransitions')



from setuptools import setup, find_packages

version_parts = (0, 8, 2)
version = '.'.join(map(str, version_parts))

setup(
    name='chromatin',
    description='neovim python plugin manager',
    version=version,
    author='Torsten Schmits',
    author_email='torstenschmits@gmail.com',
    license='MIT',
    url='https://github.com/tek/chromatin',
    include_package_data=True,
    packages=find_packages(exclude=['unit', 'unit.*', 'integration', 'integration.*']),
    install_requires=[
        'ribosome~=10.18.3',
        'pyuv',
    ],
    tests_require=[
        'kallikrein',
    ],
    entry_points={
        'console_scripts': [
            'crm_run = chromatin.cli:run',
        ],
    },
)
